// FLAC encoder plug-in for VUPlayer
// Copyright (C) 2003-2004 James Chapman
// http://www.vuplayer.com

// You will need the FLAC includes & libs to build this project.

#define FLAC__NO_DLL

#include <windows.h>
#include <commctrl.h>
#include <stdlib.h>
#include <flac/all.h>
#include "../vu_enc.h"
#include "resource.h"

extern "C" long _ftol(double);
extern "C" long _ftol2(double dblSource) {return _ftol(dblSource);}

#define MAXQUALITY	8
#define PADDING		4096

const char description[]="Free Lossless Audio Codec 1.1.2";
UINT config=5;
DWORD samplerate=44100;
WORD channels=2;
WORD bps=16;
void GetConfig();
void SetConfig();
void AddFLACTag(int type, TAGSTRUCT* pTag, FLAC__StreamMetadata* vorbiscomment);
BOOL CALLBACK DialogProc(HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

char outPath[MAX_PATH];
FLAC__int32 flacBuf[65536];
FLAC__FileEncoder* pEncoder=0;
FLAC__StreamMetadata* seektable=0;
short* shortBuf;
DWORD bufpos;

/////////////////////
// Exported functions
/////////////////////

LPCTSTR QueryEncoder()
{
	return description;
}

LPCTSTR OpenEx(LPCTSTR filename, TAGSTRUCT* pTag, DWORD dwLength, DWORD freq, WORD chan, WORD fmt)
{
	if(fmt!=FORMAT_16BIT)
		return 0;
	samplerate=freq;
	channels=chan;
	return Open(filename,pTag,dwLength);
}

LPCTSTR Open(LPCTSTR filename, TAGSTRUCT* pTag, DWORD dwLength)
{
	GetConfig();
	pEncoder=FLAC__file_encoder_new();
	if(!pEncoder)
		return 0;
	ZeroMemory(outPath,MAX_PATH);
	lstrcpy(outPath,filename);
	lstrcat(outPath,".flac");
	if(FLAC__file_encoder_set_filename(pEncoder,outPath))
	{
		FLAC__file_encoder_set_channels(pEncoder,channels);
		FLAC__file_encoder_set_sample_rate(pEncoder,samplerate);
		FLAC__file_encoder_set_bits_per_sample(pEncoder,bps);
		switch(config)
		{
		case 0:
			FLAC__file_encoder_set_blocksize(pEncoder,1152);
			FLAC__file_encoder_set_do_exhaustive_model_search(pEncoder,false);
			FLAC__file_encoder_set_do_escape_coding(pEncoder,false);
			if(channels==2)
			{
				FLAC__file_encoder_set_do_mid_side_stereo(pEncoder,false);
				FLAC__file_encoder_set_loose_mid_side_stereo(pEncoder,false);
			}
			FLAC__file_encoder_set_qlp_coeff_precision(pEncoder,0);
			FLAC__file_encoder_set_min_residual_partition_order(pEncoder,2);
			FLAC__file_encoder_set_max_residual_partition_order(pEncoder,2);
			FLAC__file_encoder_set_rice_parameter_search_dist(pEncoder,0);
			FLAC__file_encoder_set_max_lpc_order(pEncoder,0);
			break;
		case 1:
			FLAC__file_encoder_set_blocksize(pEncoder,1152);
			FLAC__file_encoder_set_do_exhaustive_model_search(pEncoder,false);
			FLAC__file_encoder_set_do_escape_coding(pEncoder,false);
			if(channels==2)
			{
				FLAC__file_encoder_set_do_mid_side_stereo(pEncoder,true);
				FLAC__file_encoder_set_loose_mid_side_stereo(pEncoder,true);
			}
			FLAC__file_encoder_set_qlp_coeff_precision(pEncoder,0);
			FLAC__file_encoder_set_min_residual_partition_order(pEncoder,2);
			FLAC__file_encoder_set_max_residual_partition_order(pEncoder,2);
			FLAC__file_encoder_set_rice_parameter_search_dist(pEncoder,0);
			FLAC__file_encoder_set_max_lpc_order(pEncoder,0);
			break;
		case 2:
			FLAC__file_encoder_set_blocksize(pEncoder,1152);
			FLAC__file_encoder_set_do_exhaustive_model_search(pEncoder,false);
			FLAC__file_encoder_set_do_escape_coding(pEncoder,false);
			if(channels==2)
			{
				FLAC__file_encoder_set_do_mid_side_stereo(pEncoder,true);
				FLAC__file_encoder_set_loose_mid_side_stereo(pEncoder,false);
			}
			FLAC__file_encoder_set_qlp_coeff_precision(pEncoder,0);
			FLAC__file_encoder_set_min_residual_partition_order(pEncoder,0);
			FLAC__file_encoder_set_max_residual_partition_order(pEncoder,3);
			FLAC__file_encoder_set_rice_parameter_search_dist(pEncoder,0);
			FLAC__file_encoder_set_max_lpc_order(pEncoder,0);
			break;
		case 3:
			FLAC__file_encoder_set_blocksize(pEncoder,4608);
			FLAC__file_encoder_set_do_exhaustive_model_search(pEncoder,false);
			FLAC__file_encoder_set_do_escape_coding(pEncoder,false);
			if(channels==2)
			{
				FLAC__file_encoder_set_do_mid_side_stereo(pEncoder,false);
				FLAC__file_encoder_set_loose_mid_side_stereo(pEncoder,false);
			}
			FLAC__file_encoder_set_qlp_coeff_precision(pEncoder,0);
			FLAC__file_encoder_set_min_residual_partition_order(pEncoder,3);
			FLAC__file_encoder_set_max_residual_partition_order(pEncoder,3);
			FLAC__file_encoder_set_rice_parameter_search_dist(pEncoder,0);
			FLAC__file_encoder_set_max_lpc_order(pEncoder,6);
			break;
		case 4:
			FLAC__file_encoder_set_blocksize(pEncoder,4608);
			FLAC__file_encoder_set_do_exhaustive_model_search(pEncoder,false);
			FLAC__file_encoder_set_do_escape_coding(pEncoder,false);
			if(channels==2)
			{
				FLAC__file_encoder_set_do_mid_side_stereo(pEncoder,true);
				FLAC__file_encoder_set_loose_mid_side_stereo(pEncoder,true);
			}
			FLAC__file_encoder_set_qlp_coeff_precision(pEncoder,0);
			FLAC__file_encoder_set_min_residual_partition_order(pEncoder,3);
			FLAC__file_encoder_set_max_residual_partition_order(pEncoder,3);
			FLAC__file_encoder_set_rice_parameter_search_dist(pEncoder,0);
			FLAC__file_encoder_set_max_lpc_order(pEncoder,8);
			break;
		case 5:
			FLAC__file_encoder_set_blocksize(pEncoder,4608);
			FLAC__file_encoder_set_do_exhaustive_model_search(pEncoder,false);
			FLAC__file_encoder_set_do_escape_coding(pEncoder,false);
			if(channels==2)
			{
				FLAC__file_encoder_set_do_mid_side_stereo(pEncoder,true);
				FLAC__file_encoder_set_loose_mid_side_stereo(pEncoder,false);
			}
			FLAC__file_encoder_set_qlp_coeff_precision(pEncoder,0);
			FLAC__file_encoder_set_min_residual_partition_order(pEncoder,3);
			FLAC__file_encoder_set_max_residual_partition_order(pEncoder,3);
			FLAC__file_encoder_set_rice_parameter_search_dist(pEncoder,0);
			FLAC__file_encoder_set_max_lpc_order(pEncoder,8);
			break;
		case 6:
			FLAC__file_encoder_set_blocksize(pEncoder,4608);
			FLAC__file_encoder_set_do_exhaustive_model_search(pEncoder,false);
			FLAC__file_encoder_set_do_escape_coding(pEncoder,false);
			if(channels==2)
			{
				FLAC__file_encoder_set_do_mid_side_stereo(pEncoder,true);
				FLAC__file_encoder_set_loose_mid_side_stereo(pEncoder,false);
			}
			FLAC__file_encoder_set_qlp_coeff_precision(pEncoder,0);
			FLAC__file_encoder_set_min_residual_partition_order(pEncoder,0);
			FLAC__file_encoder_set_max_residual_partition_order(pEncoder,4);
			FLAC__file_encoder_set_rice_parameter_search_dist(pEncoder,0);
			FLAC__file_encoder_set_max_lpc_order(pEncoder,8);
			break;
		case 7:
			FLAC__file_encoder_set_blocksize(pEncoder,4608);
			FLAC__file_encoder_set_do_exhaustive_model_search(pEncoder,true);
			FLAC__file_encoder_set_do_escape_coding(pEncoder,false);
			if(channels==2)
			{
				FLAC__file_encoder_set_do_mid_side_stereo(pEncoder,true);
				FLAC__file_encoder_set_loose_mid_side_stereo(pEncoder,false);
			}
			FLAC__file_encoder_set_qlp_coeff_precision(pEncoder,0);
			FLAC__file_encoder_set_min_residual_partition_order(pEncoder,0);
			FLAC__file_encoder_set_max_residual_partition_order(pEncoder,6);
			FLAC__file_encoder_set_rice_parameter_search_dist(pEncoder,0);
			FLAC__file_encoder_set_max_lpc_order(pEncoder,8);
			break;
		case 8:
			FLAC__file_encoder_set_blocksize(pEncoder,4608);
			FLAC__file_encoder_set_do_exhaustive_model_search(pEncoder,true);
			FLAC__file_encoder_set_do_escape_coding(pEncoder,false);
			if(channels==2)
			{
				FLAC__file_encoder_set_do_mid_side_stereo(pEncoder,true);
				FLAC__file_encoder_set_loose_mid_side_stereo(pEncoder,false);
			}
			FLAC__file_encoder_set_qlp_coeff_precision(pEncoder,0);
			FLAC__file_encoder_set_min_residual_partition_order(pEncoder,0);
			FLAC__file_encoder_set_max_residual_partition_order(pEncoder,6);
			FLAC__file_encoder_set_rice_parameter_search_dist(pEncoder,0);
			FLAC__file_encoder_set_max_lpc_order(pEncoder,12);
			break;
		}
		FLAC__StreamMetadata* metadata[3];
		seektable=FLAC__metadata_object_new(FLAC__METADATA_TYPE_SEEKTABLE);
		FLAC__StreamMetadata* vorbiscomment=FLAC__metadata_object_new(FLAC__METADATA_TYPE_VORBIS_COMMENT);
		FLAC__StreamMetadata* padding=FLAC__metadata_object_new(FLAC__METADATA_TYPE_PADDING);
		if(vorbiscomment&&seektable&&padding)
		{
			FLAC__metadata_object_seektable_template_append_spaced_points(seektable,100,dwLength/channels/(bps/8));
			for(int index=1;index<=7;index++)
				AddFLACTag(index,pTag,vorbiscomment);
			padding->length=PADDING;
			metadata[0]=vorbiscomment;
			metadata[1]=padding;
			metadata[2]=seektable;
			FLAC__file_encoder_set_metadata(pEncoder,metadata,3);
		}
		if(FLAC__file_encoder_init(pEncoder)!=FLAC__FILE_ENCODER_OK)
		{
			FLAC__file_encoder_delete(pEncoder);
			if(seektable)
			{
				FLAC__metadata_object_delete(seektable);
				seektable=0;
			}
			if(vorbiscomment)
				FLAC__metadata_object_delete(vorbiscomment);
			if(padding)
				FLAC__metadata_object_delete(padding);
			pEncoder=0;
			return 0;
		}
		if(vorbiscomment)
			FLAC__metadata_object_delete(vorbiscomment);
		if(padding)
			FLAC__metadata_object_delete(padding);
	}
	return outPath;
}

void Close()
{
	if(pEncoder)
	{
		FLAC__file_encoder_finish(pEncoder);
		FLAC__file_encoder_delete(pEncoder);
		if(seektable)
		{
			FLAC__metadata_object_delete(seektable);
			seektable=0;
		}
		pEncoder=0;
	}
}

BOOL Write(BYTE* pData, DWORD dwSize)
{
	shortBuf=(short*)pData;
	for(bufpos=0;bufpos<dwSize/2;bufpos++)
		flacBuf[bufpos]=shortBuf[bufpos];
	return FLAC__file_encoder_process_interleaved(pEncoder,flacBuf,dwSize/channels/(bps/8));
}

void Configure(HINSTANCE hInst, HWND hWnd)
{
	DialogBox(hInst,MAKEINTRESOURCE(IDD_CONFIG),hWnd,DialogProc);
}

/////////////////////
// Internal functions
/////////////////////

void GetConfig()
{
	HKEY hKey;
	if(RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer",0,KEY_QUERY_VALUE,&hKey)==ERROR_SUCCESS)
	{
		BYTE buffer[4];
		DWORD bufSize=4;
		if(RegQueryValueEx(hKey,"encoder_flac",0,0,buffer,&bufSize)==ERROR_SUCCESS)
		{
			config=*(DWORD*)buffer;
			if(config>MAXQUALITY)
				config=5;
		}
		RegCloseKey(hKey);
	}
}

void SetConfig()
{
	DWORD disp;
	HKEY hKey;
	if(RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer",0,0,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,0,&hKey,&disp)==ERROR_SUCCESS)
	{
		RegSetValueEx(hKey,"encoder_flac",0,REG_DWORD,(BYTE*)&config,4);
		RegCloseKey(hKey);
	}
}

void AddFLACTag(int type, TAGSTRUCT* pTag, FLAC__StreamMetadata* vorbiscomment)
{
	int length=0;
	switch(type)
	{
	case 1:
		length=lstrlen(pTag->artist);
		break;
	case 2:
		length=lstrlen(pTag->title);
		break;
	case 3:
		length=lstrlen(pTag->album);
		break;
	case 4:
		length=lstrlen(pTag->genre);
		break;
	case 5:
		length=lstrlen(pTag->year);
		break;
	case 6:
		length=lstrlen(pTag->comment);
		break;
	case 7:
		if(pTag->tracknum)
			length=16;
	}
	if(!length)
		return;
	char* comment=new char[length+16];
	if(comment)
	{
		switch(type)
		{
		case 1:
			lstrcpy(comment,"ARTIST=");
			lstrcat(comment,pTag->artist);
			break;
		case 2:
			lstrcpy(comment,"TITLE=");
			lstrcat(comment,pTag->title);
			break;
		case 3:
			lstrcpy(comment,"ALBUM=");
			lstrcat(comment,pTag->album);
			break;
		case 4:
			lstrcpy(comment,"GENRE=");
			lstrcat(comment,pTag->genre);
			break;
		case 5:
			lstrcpy(comment,"DATE=");
			lstrcat(comment,pTag->year);
			break;
		case 6:
			lstrcpy(comment,"COMMENT=");
			lstrcat(comment,pTag->comment);
			break;
		case 7:
			char track[16]="";
			itoa(pTag->tracknum,track,10);
			lstrcpy(comment,"TRACKNUMBER=");
			lstrcat(comment,track);
		}
		FLAC__StreamMetadata_VorbisComment_Entry commententry;
		commententry.length=lstrlen(comment);
		commententry.entry=(FLAC__byte*)comment;
		FLAC__metadata_object_vorbiscomment_insert_comment(vorbiscomment,vorbiscomment->data.vorbis_comment.num_comments,commententry,true);
		delete [] comment;
	}
}

BOOL CALLBACK DialogProc(HWND hWndDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_INITDIALOG:
		GetConfig();
		SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_SETRANGEMIN,0,0);
		SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_SETRANGEMAX,0,MAXQUALITY);
		SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_SETTICFREQ,1,0);
		SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_SETPAGESIZE,0,1);
		SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_SETPOS,1,config);
		return 1;
	case WM_CLOSE:
		PostMessage(hWndDlg,WM_COMMAND,IDCANCEL,0);
		return 1;
	case WM_NOTIFY:
		{
			LPNMHDR pnmh=(LPNMHDR)lParam;
			if(pnmh->idFrom==IDC_QUALITYSLIDER)
			{
				char buffer[2]="";
				int pos=SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_GETPOS,0,0);
				itoa(pos,buffer,10);
				SetWindowText(GetDlgItem(hWndDlg,IDC_QUALITY),buffer);
				return 1;
			}
		}
		break;
	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case IDOK:
			config=SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_GETPOS,0,0);
			SetConfig();
		case IDCANCEL:
			EndDialog(hWndDlg,0);
			return 1;
		}
 	}
	return 0;
}

