// Monkey's Audio encoder plug-in for VUPlayer
// Copyright (C) 2004 James Chapman
// http://www.vuplayer.com

// You will need the Monkey's Audio includes & libs to build this project.

#include <windows.h>
#include <ape/all.h>
#include <ape/maclib.h>
#include <ape/apetag.h>
#include "../vu_enc.h"
#include "resource.h"

extern "C" long _ftol(double);
extern "C" long _ftol2(double dblSource) {return _ftol(dblSource);}

void GetConfig();
void SetConfig();
void SetTags();
BOOL CALLBACK DialogProc(HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

const char description[]="Monkey's Audio 3.99";
DWORD samplerate=44100;
WORD channels=2;
WORD bps=16;
char outPath[MAX_PATH];
int config=1;
const int MAX_SETTING=4;
const char* SETTINGS[MAX_SETTING]={"Fast","Normal","High","Extra High"};//,"Insane"};

IAPECompress* apeCompress=0;
TAGSTRUCT* tag=0;

/////////////////////
// Exported functions
/////////////////////

LPCTSTR QueryEncoder()
{
	return description;
}

LPCTSTR OpenEx(LPCTSTR filename, TAGSTRUCT* pTag, DWORD dwLength, DWORD freq, WORD chan, WORD fmt)
{
	if(fmt!=FORMAT_16BIT)
		return 0;
	samplerate=freq;
	channels=chan;
	return Open(filename,pTag,dwLength);
}

LPCTSTR Open(LPCTSTR filename, TAGSTRUCT* pTag, DWORD dwLength)
{
	GetConfig();
	apeCompress=CreateIAPECompress();
	if(apeCompress)
	{
		WAVEFORMATEX wavFormat;
		wavFormat.cbSize=sizeof(WAVEFORMATEX);
		wavFormat.wFormatTag=WAVE_FORMAT_PCM;
		wavFormat.nSamplesPerSec=samplerate;
		wavFormat.nChannels=channels;
		wavFormat.wBitsPerSample=bps;
		wavFormat.nBlockAlign=channels*(bps/8);
		wavFormat.nAvgBytesPerSec=wavFormat.nSamplesPerSec*wavFormat.nBlockAlign;
		ZeroMemory(outPath,MAX_PATH);
		lstrcpy(outPath,filename);
		lstrcat(outPath,".ape");
		int compression;
		switch(config)
		{
		case 0:
			compression=COMPRESSION_LEVEL_FAST;
			break;
		case 1:
			compression=COMPRESSION_LEVEL_NORMAL;
			break;
		case 2:
			compression=COMPRESSION_LEVEL_HIGH;
			break;
		case 3:
			compression=COMPRESSION_LEVEL_EXTRA_HIGH;
//			break;
//		case 4:
//			compression=COMPRESSION_LEVEL_INSANE;
		}
		int wLen=MultiByteToWideChar(GetACP(),0,outPath,-1,0,0);
		if(wLen)
		{
			WCHAR* wBuf=new WCHAR[wLen];
			if(wBuf)
			{
				MultiByteToWideChar(GetACP(),0,outPath,-1,wBuf,wLen);
				if(!apeCompress->Start(wBuf,&wavFormat,dwLength,compression))
				{
					tag=pTag;
					return outPath;
				}
				delete [] wBuf;
			}
		}
		delete apeCompress;
		apeCompress=0;
	}
	return 0;
}

void Close()
{
	if(apeCompress)
	{
		apeCompress->Finish(0,0,0);
		delete apeCompress;
		SetTags();
	}
}

BOOL Write(BYTE* pData, DWORD dwSize)
{
	if(apeCompress->AddData(pData,dwSize))
		return 0;
	else
		return 1;
}

void Configure(HINSTANCE hInst, HWND hWnd)
{
	DialogBox(hInst,MAKEINTRESOURCE(IDD_CONFIG),hWnd,DialogProc);
}

/////////////////////
// Internal functions
/////////////////////

void GetConfig()
{
	HKEY hKey;
	if(RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer",0,KEY_QUERY_VALUE,&hKey)==ERROR_SUCCESS)
	{
		BYTE buffer[4];
		DWORD bufSize=4;
		if(RegQueryValueEx(hKey,"encoder_mac",0,0,buffer,&bufSize)==ERROR_SUCCESS)
		{
			config=*(DWORD*)buffer;
			if(config>=MAX_SETTING)
				config=1;
		}
		RegCloseKey(hKey);
	}
}

void SetConfig()
{
	DWORD disp;
	HKEY hKey;
	if(RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer",0,0,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,0,&hKey,&disp)==ERROR_SUCCESS)
	{
		RegSetValueEx(hKey,"encoder_mac",0,REG_DWORD,(BYTE*)&config,4);
		RegCloseKey(hKey);
	}
}

void SetTags()
{
	int wLen=MultiByteToWideChar(GetACP(),0,outPath,-1,0,0);
	if(wLen)
	{
		WCHAR* wBuf=new WCHAR[wLen];
		if(wBuf)
		{
			MultiByteToWideChar(GetACP(),0,outPath,-1,wBuf,wLen);
			CAPETag apetag(wBuf);
			apetag.SetFieldString(APE_TAG_FIELD_TITLE,tag->title,1);
			apetag.SetFieldString(APE_TAG_FIELD_ARTIST,tag->artist,1);
			apetag.SetFieldString(APE_TAG_FIELD_ALBUM,tag->album,1);
			apetag.SetFieldString(APE_TAG_FIELD_GENRE,tag->genre,1);
			apetag.SetFieldString(APE_TAG_FIELD_YEAR,tag->year,1);
			apetag.SetFieldString(APE_TAG_FIELD_COMMENT,tag->comment,1);
			char tracknum[32]="";
			itoa(tag->tracknum,tracknum,10);
			apetag.SetFieldString(APE_TAG_FIELD_TRACK,tracknum,1);
			apetag.Save();
			delete [] wBuf;
		}
	}
}

BOOL CALLBACK DialogProc(HWND hWndDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_INITDIALOG:
		{
			GetConfig();
			int index;
			for(index=0;index<MAX_SETTING;index++)
				SendDlgItemMessage(hWndDlg,IDC_SETTING,CB_ADDSTRING,0,(LPARAM)SETTINGS[index]);
			SendDlgItemMessage(hWndDlg,IDC_SETTING,CB_SETCURSEL,config,0);
		}
		return 1;
	case WM_CLOSE:
		PostMessage(hWndDlg,WM_COMMAND,IDCANCEL,0);
		return 1;
	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case IDOK:
			config=SendDlgItemMessage(hWndDlg,IDC_SETTING,CB_GETCURSEL,0,0);
			SetConfig();
		case IDCANCEL:
			EndDialog(hWndDlg,0);
			return 1;
		}
 	}
	return 0;
}



