// LAME MP3 encoder plug-in for VUPlayer
// Copyright (C) 2004 James Chapman
// http://www.vuplayer.com

// You will need the LAME includes & libs to build this project.

#include <windows.h>
#include <lame/lame.h>
#include "../vu_enc.h"
#include "resource.h"

extern "C" long _ftol(double);
extern "C" long _ftol2(double dblSource) {return _ftol(dblSource);}

#define BUFFERSIZE		65536

const int COUNT_CBR=14;
const int COUNT_VBR=10;
const int COUNT_PRESET=4;
const int SETTINGS_CBR[COUNT_CBR]={32,40,48,56,64,80,96,112,128,160,192,224,256,320};
const int SETTINGS_VBR[COUNT_VBR]={9,8,7,6,5,4,3,2,1,0};
const char* SETTINGS_PRESET[COUNT_PRESET]={"Medium","Standard","Extreme","Insane"};
const int MAX_SETTING=COUNT_CBR+COUNT_VBR+COUNT_PRESET;
const int DEFAULT_SETTING=25;

const char CBR_PREFIX[]="CBR ";
const char CBR_SUFFIX[]=" kbps";
const char VBR_PREFIX[]="VBR Quality ";
const char PRESET_PREFIX[]="Preset ";

char description[32]="LAME ";
int config=DEFAULT_SETTING;
DWORD samplerate=44100;
WORD channels=2;
WORD bps=16;
char* ansi_artist=0;
char* ansi_title=0;
char* ansi_album=0;
char* ansi_comment=0;
void GetConfig();
void SetConfig();
void ClearTags();
BOOL CALLBACK DialogProc(HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

lame_global_flags* gfp;
FILE* outFile;
char outPath[MAX_PATH];
unsigned char mp3Buffer[BUFFERSIZE];

/////////////////////
// Exported functions
/////////////////////

LPCTSTR QueryEncoder()
{
	lstrcat(description,get_lame_version());
	return description;
}

LPCTSTR OpenEx(LPCTSTR filename, TAGSTRUCT* pTag, DWORD dwLength, DWORD freq, WORD chan, WORD fmt)
{
	if(fmt!=FORMAT_16BIT)
		return 0;
	samplerate=freq;
	channels=chan;
	return Open(filename,pTag,dwLength);
}

LPCTSTR Open(LPCTSTR filename, TAGSTRUCT* pTag, DWORD dwLength)
{
	GetConfig();
	gfp=lame_init();
	if(gfp)
	{
		lame_set_quality(gfp,2);
		lame_set_num_channels(gfp,channels);
		lame_set_in_samplerate(gfp,samplerate);
		if(config<COUNT_CBR)
			lame_set_brate(gfp,SETTINGS_CBR[config]);
		else
		{
			if(config<(COUNT_CBR+COUNT_VBR))
			{
				lame_set_VBR(gfp,vbr_default);
				lame_set_VBR_q(gfp,SETTINGS_VBR[config-COUNT_CBR]);
			}
			else
			{
				switch(config)
				{
				case COUNT_CBR+COUNT_VBR:
					lame_set_preset(gfp,MEDIUM);
					break;
				case COUNT_CBR+COUNT_VBR+1:
					lame_set_preset(gfp,STANDARD);
					break;
				case COUNT_CBR+COUNT_VBR+2:
					lame_set_preset(gfp,EXTREME);
					break;
				case COUNT_CBR+COUNT_VBR+3:
					lame_set_preset(gfp,INSANE);
				}
			}
		}
		id3tag_v1_only(gfp);
		// Convert UTF-8 back to ANSI for ID3v1 tags
		int length=MultiByteToWideChar(CP_UTF8,0,pTag->title,-1,0,0);
		if(length)
		{
			WCHAR* wBuf=new WCHAR[length];
			if(wBuf)
			{
				MultiByteToWideChar(CP_UTF8,0,pTag->title,-1,wBuf,length);
				length=WideCharToMultiByte(GetACP(),0,wBuf,-1,0,0,0,0);
				if(length)
				{
					ansi_title=new char[length];
					if(ansi_title)
					{
						WideCharToMultiByte(GetACP(),0,wBuf,-1,ansi_title,length,0,0);
						id3tag_set_title(gfp,ansi_title);
					}
				}
				delete [] wBuf;
			}
		}
		length=MultiByteToWideChar(CP_UTF8,0,pTag->artist,-1,0,0);
		if(length)
		{
			WCHAR* wBuf=new WCHAR[length];
			if(wBuf)
			{
				MultiByteToWideChar(CP_UTF8,0,pTag->artist,-1,wBuf,length);
				length=WideCharToMultiByte(GetACP(),0,wBuf,-1,0,0,0,0);
				if(length)
				{
					ansi_artist=new char[length];
					if(ansi_artist)
					{
						WideCharToMultiByte(GetACP(),0,wBuf,-1,ansi_artist,length,0,0);
						id3tag_set_artist(gfp,ansi_artist);
					}
				}
				delete [] wBuf;
			}
		}
		length=MultiByteToWideChar(CP_UTF8,0,pTag->album,-1,0,0);
		if(length)
		{
			WCHAR* wBuf=new WCHAR[length];
			if(wBuf)
			{
				MultiByteToWideChar(CP_UTF8,0,pTag->album,-1,wBuf,length);
				length=WideCharToMultiByte(GetACP(),0,wBuf,-1,0,0,0,0);
				if(length)
				{
					ansi_album=new char[length];
					if(ansi_album)
					{
						WideCharToMultiByte(GetACP(),0,wBuf,-1,ansi_album,length,0,0);
						id3tag_set_album(gfp,ansi_album);
					}
				}
				delete [] wBuf;
			}
		}
		length=MultiByteToWideChar(CP_UTF8,0,pTag->comment,-1,0,0);
		if(length)
		{
			WCHAR* wBuf=new WCHAR[length];
			if(wBuf)
			{
				MultiByteToWideChar(CP_UTF8,0,pTag->comment,-1,wBuf,length);
				length=WideCharToMultiByte(GetACP(),0,wBuf,-1,0,0,0,0);
				if(length)
				{
					ansi_comment=new char[length];
					if(ansi_comment)
					{
						WideCharToMultiByte(GetACP(),0,wBuf,-1,ansi_comment,length,0,0);
						id3tag_set_comment(gfp,ansi_comment);
					}
				}
				delete [] wBuf;
			}
		}
		id3tag_set_year(gfp,pTag->year);
		char tracknum[16]="";
		itoa(pTag->tracknum,tracknum,10);
		id3tag_set_track(gfp,tracknum);
		if(lstrlen(pTag->genre))
		{
			if(id3tag_set_genre(gfp,pTag->genre))
				id3tag_set_genre(gfp,"Other");
		}
		else
			id3tag_set_genre(gfp,"Other");
		if(lame_init_params(gfp)>=0)
		{
			ZeroMemory(outPath,MAX_PATH);
			lstrcpy(outPath,filename);
			lstrcat(outPath,".mp3");
			outFile=fopen(outPath,"w+b");
			if(outFile)
				return outPath;
		}
		lame_close(gfp);
	}
	ClearTags();
	return 0;
}

void Close()
{
	int bytes=lame_encode_flush(gfp,mp3Buffer,BUFFERSIZE);
	if(bytes>0)
		fwrite(mp3Buffer,1,bytes,outFile);
	lame_mp3_tags_fid(gfp,outFile);
	lame_close(gfp);
	fclose(outFile);
	ClearTags();
}

BOOL Write(BYTE* pData, DWORD dwSize)
{
	int bytes;
	if(channels==2)
		bytes=lame_encode_buffer_interleaved(gfp,(short*)pData,dwSize/4,mp3Buffer,BUFFERSIZE);
	else
		bytes=lame_encode_buffer(gfp,(short*)pData,0,dwSize/2,mp3Buffer,BUFFERSIZE);
	if(bytes>=0)
	{
		fwrite(mp3Buffer,1,bytes,outFile);
		return 1;
	}
	else
		return 0;
}

void Configure(HINSTANCE hInst, HWND hWnd)
{
	DialogBox(hInst,MAKEINTRESOURCE(IDD_CONFIG),hWnd,DialogProc);
}

/////////////////////
// Internal functions
/////////////////////

void GetConfig()
{
	HKEY hKey;
	if(RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer",0,KEY_QUERY_VALUE,&hKey)==ERROR_SUCCESS)
	{
		BYTE buffer[4];
		DWORD bufSize=4;
		if(RegQueryValueEx(hKey,"encoder_mp3",0,0,buffer,&bufSize)==ERROR_SUCCESS)
		{
			config=*(DWORD*)buffer;
			if(config>=MAX_SETTING)
				config=DEFAULT_SETTING;
		}
		RegCloseKey(hKey);
	}
}

void SetConfig()
{
	DWORD disp;
	HKEY hKey;
	if(RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer",0,0,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,0,&hKey,&disp)==ERROR_SUCCESS)
	{
		RegSetValueEx(hKey,"encoder_mp3",0,REG_DWORD,(BYTE*)&config,4);
		RegCloseKey(hKey);
	}
}

void ClearTags()
{
	if(ansi_artist)
	{
		delete [] ansi_artist;
		ansi_artist=0;
	}
	if(ansi_title)
	{
		delete [] ansi_title;
		ansi_title=0;
	}
	if(ansi_album)
	{
		delete [] ansi_album;
		ansi_album=0;
	}
	if(ansi_comment)
	{
		delete [] ansi_comment;
		ansi_comment=0;
	}
}

BOOL CALLBACK DialogProc(HWND hWndDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_INITDIALOG:
		{
			GetConfig();
			int index;
			char entry[32]="";
			char buffer[8]="";
			for(index=0;index<COUNT_CBR;index++)
			{
				lstrcpy(entry,CBR_PREFIX);
				itoa(SETTINGS_CBR[index],buffer,10);
				lstrcat(entry,buffer);
				lstrcat(entry,CBR_SUFFIX);
				SendDlgItemMessage(hWndDlg,IDC_SETTING,CB_ADDSTRING,0,(LPARAM)entry);
			}
			for(index=COUNT_CBR;index<(COUNT_CBR+COUNT_VBR);index++)
			{
				lstrcpy(entry,VBR_PREFIX);
				itoa(SETTINGS_VBR[index-COUNT_CBR],buffer,10);
				lstrcat(entry,buffer);
				SendDlgItemMessage(hWndDlg,IDC_SETTING,CB_ADDSTRING,0,(LPARAM)entry);
			}
			for(index=COUNT_CBR+COUNT_VBR;index<(COUNT_CBR+COUNT_VBR+COUNT_PRESET);index++)
			{
				lstrcpy(entry,PRESET_PREFIX);
				lstrcat(entry,SETTINGS_PRESET[index-COUNT_CBR-COUNT_VBR]);
				SendDlgItemMessage(hWndDlg,IDC_SETTING,CB_ADDSTRING,0,(LPARAM)entry);
			}
			SendDlgItemMessage(hWndDlg,IDC_SETTING,CB_SETCURSEL,config,0);
		}
		return 1;
	case WM_CLOSE:
		PostMessage(hWndDlg,WM_COMMAND,IDCANCEL,0);
		return 1;
	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case IDOK:
			config=SendDlgItemMessage(hWndDlg,IDC_SETTING,CB_GETCURSEL,0,0);
			SetConfig();
		case IDCANCEL:
			EndDialog(hWndDlg,0);
			return 1;
		}
 	}
	return 0;
}
