// Ogg Vorbis encoder plug-in for VUPlayer
// Copyright (C) 2003-2004 James Chapman
// http://www.vuplayer.com

// You will need the Ogg Vorbis includes & libs to build this project.
// For simplicity, this example just uses VBR encoding.

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <time.h>
#include <vorbis/vorbisenc.h>
#include "../vu_enc.h"
#include "resource.h"

extern "C" long _ftol(double);
extern "C" long _ftol2(double dblSource) {return _ftol(dblSource);}

#define MAXQUALITY		110
#define VORBIS_BLOCK	32768

const char description[]="Ogg Vorbis 1.1.0";
int config=50;
DWORD samplerate=44100;
WORD channels=2;
WORD bps=16;
void GetConfig();
void SetConfig();
void AddOGGTag(int type, TAGSTRUCT* pTag);
BOOL CALLBACK DialogProc(HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

FILE* outFile;
char outPath[MAX_PATH];
ogg_stream_state os;
ogg_page         og;
ogg_packet       op;
vorbis_info      vi;
vorbis_comment   vc;
vorbis_dsp_state vd;
vorbis_block     vb;

/////////////////////
// Exported functions
/////////////////////

LPCTSTR QueryEncoder()
{
	return description;
}

LPCTSTR OpenEx(LPCTSTR filename, TAGSTRUCT* pTag, DWORD dwLength, DWORD freq, WORD chan, WORD fmt)
{
	samplerate=freq;
	channels=chan;
	if(fmt==FORMAT_FLOAT)
		bps=32;
	return Open(filename,pTag,dwLength);
}

LPCTSTR Open(LPCTSTR filename, TAGSTRUCT* pTag, DWORD dwLength)
{
	GetConfig();
	ZeroMemory(outPath,MAX_PATH);
	lstrcpy(outPath,filename);
	lstrcat(outPath,".ogg");
	outFile=fopen(outPath,"wb");
	if(outFile)
	{
		float fQual=(float)(config-10);
		fQual/=100.0f;
		vorbis_info_init(&vi);
		vorbis_encode_init_vbr(&vi,channels,samplerate,fQual);
		vorbis_comment_init(&vc);
		for(int index=1;index<=7;index++)
			AddOGGTag(index,pTag);
		vorbis_analysis_init(&vd,&vi);
		vorbis_block_init(&vd,&vb);
		srand(time(0));
		ogg_stream_init(&os,rand());
		ogg_packet header;
		ogg_packet header_comm;
		ogg_packet header_code;
		vorbis_analysis_headerout(&vd,&vc,&header,&header_comm,&header_code);
		ogg_stream_packetin(&os,&header);
		ogg_stream_packetin(&os,&header_comm);
		ogg_stream_packetin(&os,&header_code);
		int result=ogg_stream_flush(&os,&og);
		while(result)
		{
			fwrite(og.header,1,og.header_len,outFile);
			fwrite(og.body,1,og.body_len,outFile);
			result=ogg_stream_flush(&os,&og);
		}
		return outPath;
	}
	else
		return 0;
}

void Close()
{
	vorbis_analysis_wrote(&vd,0);
	while(vorbis_analysis_blockout(&vd,&vb)==1)
	{
		vorbis_analysis(&vb,0);
		vorbis_bitrate_addblock(&vb);
		while(vorbis_bitrate_flushpacket(&vd,&op))
		{
			ogg_stream_packetin(&os,&op);
			int result=ogg_stream_pageout(&os,&og);
			while(result)
			{
				fwrite(og.header,1,og.header_len,outFile);
				fwrite(og.body,1,og.body_len,outFile);
				if(ogg_page_eos(&og))
					break;
				result=ogg_stream_pageout(&os,&og);
			}
		}
	}
	ogg_stream_clear(&os);
	vorbis_block_clear(&vb);
	vorbis_dsp_clear(&vd);
	vorbis_comment_clear(&vc);
	vorbis_info_clear(&vi);
	fclose(outFile);
}

BOOL Write(BYTE* pData, DWORD dwSize)
{
	float **buffer=vorbis_analysis_buffer(&vd,VORBIS_BLOCK);
	DWORD i,j;
	if(bps==16)
	{
		short* buf=(short*)pData;
		for(i=0;i<(dwSize/channels/(bps/8));i++)
		{
			for(j=0;j<channels;j++)
				buffer[j][i]=buf[i*channels+j]/32768.0f;
		}
	}
	else
	{
		float* buf=(float*)pData;
		for(i=0;i<(dwSize/channels/(bps/8));i++)
		{
			for(j=0;j<channels;j++)
				buffer[j][i]=buf[i*channels+j];
		}
	}
	vorbis_analysis_wrote(&vd,i);
	while(vorbis_analysis_blockout(&vd,&vb)==1)
	{
		vorbis_analysis(&vb,0);
		vorbis_bitrate_addblock(&vb);
		while(vorbis_bitrate_flushpacket(&vd,&op))
		{
			ogg_stream_packetin(&os,&op);
			int result=ogg_stream_pageout(&os,&og);
			while(result)
			{
				fwrite(og.header,1,og.header_len,outFile);
				fwrite(og.body,1,og.body_len,outFile);
				if(ogg_page_eos(&og))
					break;
				result=ogg_stream_pageout(&os,&og);
			}
		}
	}
	return 1;
}

void Configure(HINSTANCE hInst, HWND hWnd)
{
	DialogBox(hInst,MAKEINTRESOURCE(IDD_CONFIG),hWnd,DialogProc);
}

/////////////////////
// Internal functions
/////////////////////

void GetConfig()
{
	HKEY hKey;
	if(RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer",0,KEY_QUERY_VALUE,&hKey)==ERROR_SUCCESS)
	{
		BYTE buffer[4];
		DWORD bufSize=4;
		if(RegQueryValueEx(hKey,"encoder_ogg",0,0,buffer,&bufSize)==ERROR_SUCCESS)
		{
			config=*(DWORD*)buffer;
			if(config>MAXQUALITY)
				config=50;
		}
		RegCloseKey(hKey);
	}
}

void SetConfig()
{
	DWORD disp;
	HKEY hKey;
	if(RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer",0,0,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,0,&hKey,&disp)==ERROR_SUCCESS)
	{
		RegSetValueEx(hKey,"encoder_ogg",0,REG_DWORD,(BYTE*)&config,4);
		RegCloseKey(hKey);
	}
}

void AddOGGTag(int type, TAGSTRUCT* pTag)
{
	int length=0;
	switch(type)
	{
	case 1:
		length=lstrlen(pTag->artist);
		break;
	case 2:
		length=lstrlen(pTag->title);
		break;
	case 3:
		length=lstrlen(pTag->album);
		break;
	case 4:
		length=lstrlen(pTag->genre);
		break;
	case 5:
		length=lstrlen(pTag->year);
		break;
	case 6:
		length=lstrlen(pTag->comment);
		break;
	case 7:
		if(pTag->tracknum)
			length=16;
	}
	if(!length)
		return;
	char* tag=new char[length+16];
	if(tag)
	{
		switch(type)
		{
		case 1:
			lstrcpy(tag,"ARTIST=");
			lstrcat(tag,pTag->artist);
			break;
		case 2:
			lstrcpy(tag,"TITLE=");
			lstrcat(tag,pTag->title);
			break;
		case 3:
			lstrcpy(tag,"ALBUM=");
			lstrcat(tag,pTag->album);
			break;
		case 4:
			lstrcpy(tag,"GENRE=");
			lstrcat(tag,pTag->genre);
			break;
		case 5:
			lstrcpy(tag,"DATE=");
			lstrcat(tag,pTag->year);
			break;
		case 6:
			lstrcpy(tag,"COMMENT=");
			lstrcat(tag,pTag->comment);
			break;
		case 7:
			char tracknum[16]="";
			itoa(pTag->tracknum,tracknum,10);
			lstrcpy(tag,"TRACKNUMBER=");
			lstrcat(tag,tracknum);
		}
		vorbis_comment_add(&vc,tag);
		delete [] tag;
	}
}

BOOL CALLBACK DialogProc(HWND hWndDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_INITDIALOG:
		GetConfig();
		SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_SETRANGEMIN,0,0);
		SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_SETRANGEMAX,0,MAXQUALITY);
		SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_SETTICFREQ,10,0);
		SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_SETPAGESIZE,0,10);
		SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_SETPOS,1,config);
		return 1;
	case WM_CLOSE:
		PostMessage(hWndDlg,WM_COMMAND,IDCANCEL,0);
		return 1;
	case WM_NOTIFY:
		{
			LPNMHDR pnmh=(LPNMHDR)lParam;
			if(pnmh->idFrom==IDC_QUALITYSLIDER)
			{
				char buffer[8]="";
				int pos=SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_GETPOS,0,0);
				if(pos<10)
				{
					lstrcat(buffer,"-");
					pos=10-pos;
				}
				else
					pos-=10;
				buffer[lstrlen(buffer)]=(char)(pos/100+'0');
				buffer[lstrlen(buffer)]='.';
				pos%=100;
				buffer[lstrlen(buffer)]=(char)(pos/10+'0');
				buffer[lstrlen(buffer)]=(char)(pos%10+'0');
				SetWindowText(GetDlgItem(hWndDlg,IDC_QUALITY),buffer);
				return 1;
			}
		}
		break;
	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case IDOK:
			config=SendDlgItemMessage(hWndDlg,IDC_QUALITYSLIDER,TBM_GETPOS,0,0);
			SetConfig();
		case IDCANCEL:
			EndDialog(hWndDlg,0);
			return 1;
		}
 	}
	return 0;
}
