// Windows Media Audio encoder plug-in for VUPlayer
// Copyright (C) 2003-2005 James Chapman
// http://www.vuplayer.com

// You will need the Windows Media Format SDK to build this project.

#include <windows.h>
#include <commctrl.h>
#include <wmsdk/wmsdk.h>
#include "../vu_enc.h"
#include "resource.h"

#define FLAG_VBR	0x80000000

const char description[]="Windows Media Audio";
UINT config=0;
DWORD samplerate=44100;
WORD channels=2;
WORD bps=16;
void GetConfig();
void SetConfig();
void SetWMATags(TAGSTRUCT* pTag);
BOOL CALLBACK DialogProc(HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
void UpdateCodecCombo(HWND hWndDlg, BOOL vbr, DWORD codecindex);
void UpdateFormatCombo(HWND hWndDlg, BOOL vbr, DWORD codecindex);

char outPath[MAX_PATH];
IWMWriter* pWriter=0;
INSSBuffer* insBuffer=0;
QWORD cnsCurrent=0;
QWORD qwRate;
QWORD qwTemp;
BYTE* pData=0;
DWORD cbData=0;
int curVBR=0;
int curCodec=0;

/////////////////////
// Exported functions
/////////////////////

LPCTSTR QueryEncoder()
{
	return description;
}

LPCTSTR OpenEx(LPCTSTR filename, TAGSTRUCT* pTag, DWORD dwLength, DWORD freq, WORD chan, WORD fmt)
{
	if(fmt!=FORMAT_16BIT)
		return 0;
	samplerate=freq;
	channels=chan;
	return Open(filename,pTag,dwLength);
}

LPCTSTR Open(LPCTSTR filename, TAGSTRUCT* pTag, DWORD dwLength)
{
	BOOL success=0;
	GetConfig();
	cnsCurrent=0;
	pData=0;
	cbData=0;
	HRESULT hr=WMCreateWriter(0,&pWriter);
	if(SUCCEEDED(hr))
	{
		IWMProfileManager* pProfileManager;
		hr=WMCreateProfileManager(&pProfileManager);
		if(SUCCEEDED(hr))
		{
			IWMProfile* pProfile;
			hr=pProfileManager->CreateEmptyProfile(WMT_VER_9_0,&pProfile);
			if(SUCCEEDED(hr))
			{
				IWMCodecInfo3* pCodecInfo;
				hr=pProfileManager->QueryInterface(IID_IWMCodecInfo3,(void**)&pCodecInfo);
				if(SUCCEEDED(hr))
				{
					DWORD codecindex=(config>>16)&0xff;
					DWORD formatindex=config&0xffff;
					WMT_ATTR_DATATYPE datatype=WMT_TYPE_BOOL;
					BOOL vbr=0;
					if(config&FLAG_VBR)
						vbr=1;
					hr=pCodecInfo->SetCodecEnumerationSetting(WMMEDIATYPE_Audio,codecindex,g_wszVBREnabled,datatype,(LPBYTE)&vbr,sizeof(BOOL));
					if(SUCCEEDED(hr))
					{
						IWMStreamConfig* pStreamConfig;
						hr=pCodecInfo->GetCodecFormat(WMMEDIATYPE_Audio,codecindex,formatindex,&pStreamConfig);
						if(SUCCEEDED(hr))
						{
							hr=pStreamConfig->SetStreamNumber(1);
							if(SUCCEEDED(hr))
							{
								hr=pProfile->AddStream(pStreamConfig);
								if(SUCCEEDED(hr))
								{
									hr=pWriter->SetProfile(pProfile);
									if(SUCCEEDED(hr))
									{
										IWMInputMediaProps* pProps;
										hr=pWriter->GetInputProps(0,&pProps);
										if(SUCCEEDED(hr))
										{
											DWORD dwSize=0;
											hr=pProps->GetMediaType(0,&dwSize);
											if(SUCCEEDED(hr)&&dwSize)
											{
												BYTE* pInput=new BYTE[dwSize];
												if(pInput)
												{
													hr=pProps->GetMediaType((WM_MEDIA_TYPE*)pInput,&dwSize);
													if(SUCCEEDED(hr))
													{
														WM_MEDIA_TYPE* pMediaType=(WM_MEDIA_TYPE*)pInput;
														WAVEFORMATEX* pwfx=(WAVEFORMATEX*)pMediaType->pbFormat;
														pwfx->cbSize=sizeof(WAVEFORMATEX);
														pwfx->wFormatTag=WAVE_FORMAT_PCM;
														pwfx->nChannels=channels;
														pwfx->nSamplesPerSec=samplerate;
														pwfx->wBitsPerSample=bps;
														pwfx->nBlockAlign=pwfx->nChannels*(pwfx->wBitsPerSample/8);
														pwfx->nAvgBytesPerSec=pwfx->nSamplesPerSec*pwfx->nBlockAlign;
														qwRate=pwfx->nAvgBytesPerSec;
														hr=pProps->SetMediaType(pMediaType);
														if(SUCCEEDED(hr))
														{
															hr=pWriter->SetInputProps(0,pProps);
															if(SUCCEEDED(hr))
															{
																ZeroMemory(outPath,MAX_PATH);
																strcpy(outPath,filename);
																strcat(outPath,".wma");
																WCHAR output[MAX_PATH];
																ZeroMemory(&output,MAX_PATH*sizeof(WCHAR));
																MultiByteToWideChar(GetACP(),MB_PRECOMPOSED,outPath,-1,output,MAX_PATH);
																hr=pWriter->SetOutputFilename(output);
																if(SUCCEEDED(hr))
																{
																	SetWMATags(pTag);
																	hr=pWriter->BeginWriting();
																	if(SUCCEEDED(hr))
																		success=1;
																}
															}
														}
													}
													delete [] pInput;
												}
											}
											pProps->Release();
										}
									}
								}
							}
							pStreamConfig->Release();
						}
					}
					pCodecInfo->Release();
				}
				pProfile->Release();
			}
			pProfileManager->Release();
		}
	}
	if(success)
		return outPath;
	else
	{
		if(pWriter)
		{
			pWriter->Release();
			pWriter=0;
		}
		return 0;
	}
}

void Close()
{
	if(pWriter)
	{
		pWriter->EndWriting();
		pWriter->Release();
		pWriter=0;
	}
}

BOOL Write(BYTE* pBuffer, DWORD dwSize)
{
	HRESULT hr=pWriter->AllocateSample(dwSize,&insBuffer);
	if(FAILED(hr))
		return 0;
	hr=insBuffer->GetBufferAndLength(&pData,&cbData);
	if(FAILED(hr))
	{
		insBuffer->Release();
		return 0;
	}
	memcpy(pData,pBuffer,dwSize);
	hr=insBuffer->SetLength(dwSize);
	if(FAILED(hr))
	{
		insBuffer->Release();
		return 0;
	}
	hr=pWriter->WriteSample(0,cnsCurrent,0,insBuffer);
	if(insBuffer)
	{
		insBuffer->Release();
		insBuffer=0;
	}
	qwTemp=cbData;
	qwTemp*=10000000;
	qwTemp/=qwRate;
	cnsCurrent+=qwTemp;
	return 1;
}

void Configure(HINSTANCE hInst, HWND hWnd)
{
	DialogBox(hInst,MAKEINTRESOURCE(IDD_CONFIG),hWnd,DialogProc);
}

/////////////////////
// Internal functions
/////////////////////

void GetConfig()
{
	HKEY hKey;
	if(RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer",0,KEY_QUERY_VALUE,&hKey)==ERROR_SUCCESS)
	{
		BYTE buffer[4];
		DWORD bufSize=4;
		if(RegQueryValueEx(hKey,"encoder_wma2",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			config=*(DWORD*)buffer;
		RegCloseKey(hKey);
	}
}

void SetConfig()
{
	DWORD disp;
	HKEY hKey;
	if(RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer",0,0,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,0,&hKey,&disp)==ERROR_SUCCESS)
	{
		RegSetValueEx(hKey,"encoder_wma2",0,REG_DWORD,(BYTE*)&config,4);
		RegCloseKey(hKey);
	}
}

void SetWMATags(TAGSTRUCT* pTag)
{
	IWMHeaderInfo* pHeader=0;
	HRESULT hr=pWriter->QueryInterface(IID_IWMHeaderInfo,(void**)&pHeader);
	if(FAILED(hr))
		return;
	WCHAR* pBuffer=0;
	int wLength=MultiByteToWideChar(CP_UTF8,0,pTag->title,-1,0,0);
	if(wLength>1)
	{
		pBuffer=new WCHAR[wLength];
		if(pBuffer)
		{
			MultiByteToWideChar(CP_UTF8,0,pTag->title,-1,pBuffer,wLength);
			hr=pHeader->SetAttribute(0,g_wszWMTitle,WMT_TYPE_STRING,(LPBYTE)pBuffer,wLength*sizeof(WCHAR));
			delete [] pBuffer;
		}
	}
	wLength=MultiByteToWideChar(CP_UTF8,0,pTag->artist,-1,0,0);
	if(wLength>1)
	{
		pBuffer=new WCHAR[wLength];
		if(pBuffer)
		{
			MultiByteToWideChar(CP_UTF8,0,pTag->artist,-1,pBuffer,wLength);
			hr=pHeader->SetAttribute(0,g_wszWMAuthor,WMT_TYPE_STRING,(LPBYTE)pBuffer,wLength*sizeof(WCHAR));
			delete [] pBuffer;
		}
	}
	wLength=MultiByteToWideChar(CP_UTF8,0,pTag->album,-1,0,0);
	if(wLength>1)
	{
		pBuffer=new WCHAR[wLength];
		if(pBuffer)
		{
			MultiByteToWideChar(CP_UTF8,0,pTag->album,-1,pBuffer,wLength);
			hr=pHeader->SetAttribute(0,g_wszWMAlbumTitle,WMT_TYPE_STRING,(LPBYTE)pBuffer,wLength*sizeof(WCHAR));
			delete [] pBuffer;
		}
	}
	wLength=MultiByteToWideChar(CP_UTF8,0,pTag->genre,-1,0,0);
	if(wLength>1)
	{
		pBuffer=new WCHAR[wLength];
		if(pBuffer)
		{
			MultiByteToWideChar(CP_UTF8,0,pTag->genre,-1,pBuffer,wLength);
			hr=pHeader->SetAttribute(0,g_wszWMGenre,WMT_TYPE_STRING,(LPBYTE)pBuffer,wLength*sizeof(WCHAR));
			delete [] pBuffer;
		}
	}
	wLength=MultiByteToWideChar(CP_UTF8,0,pTag->year,-1,0,0);
	if(wLength>1)
	{
		pBuffer=new WCHAR[wLength];
		if(pBuffer)
		{
			MultiByteToWideChar(CP_UTF8,0,pTag->year,-1,pBuffer,wLength);
			hr=pHeader->SetAttribute(0,g_wszWMYear,WMT_TYPE_STRING,(LPBYTE)pBuffer,wLength*sizeof(WCHAR));
			delete [] pBuffer;
		}
	}
	wLength=MultiByteToWideChar(CP_UTF8,0,pTag->comment,-1,0,0);
	if(wLength>1)
	{
		pBuffer=new WCHAR[wLength];
		if(pBuffer)
		{
			MultiByteToWideChar(CP_UTF8,0,pTag->comment,-1,pBuffer,wLength);
			hr=pHeader->SetAttribute(0,g_wszWMDescription,WMT_TYPE_STRING,(LPBYTE)pBuffer,wLength*sizeof(WCHAR));
			delete [] pBuffer;
		}
	}
	DWORD track=pTag->tracknum-1;
	hr=pHeader->SetAttribute(0,g_wszWMTrack,WMT_TYPE_DWORD,(LPBYTE)&track,sizeof(DWORD));
	pHeader->Release();
}

void UpdateCodecCombo(HWND hWndDlg, BOOL vbr, DWORD codecindex)
{
	SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_RESETCONTENT,0,0);
	IWMProfileManager* pProfileManager;
	HRESULT hr=WMCreateProfileManager(&pProfileManager);
	if(SUCCEEDED(hr))
	{
		IWMCodecInfo3* pCodecInfo;
		hr=pProfileManager->QueryInterface(IID_IWMCodecInfo3,(void**)&pCodecInfo);
		if(SUCCEEDED(hr))
		{
			DWORD cCodecs=0;
			WCHAR* pwszCodecName=0;
			DWORD cchCodecName=0;
			char* pszCombo=0;
			DWORD cCombo;
			hr=pCodecInfo->GetCodecInfoCount(WMMEDIATYPE_Audio,&cCodecs);
			if(SUCCEEDED(hr))
			{
				WMT_ATTR_DATATYPE datatype=WMT_TYPE_BOOL;
				DWORD cFormats=0;
				for(DWORD dwCodecIndex=0;dwCodecIndex<cCodecs;dwCodecIndex++)
			    {
					hr=pCodecInfo->SetCodecEnumerationSetting(WMMEDIATYPE_Audio,dwCodecIndex,g_wszVBREnabled,datatype,(LPBYTE)&vbr,sizeof(BOOL));
					if(SUCCEEDED(hr))
					{
						hr=pCodecInfo->GetCodecFormatCount(WMMEDIATYPE_Audio,dwCodecIndex,&cFormats);
						if(SUCCEEDED(hr)&&cFormats)
						{
							hr=pCodecInfo->GetCodecName(WMMEDIATYPE_Audio,dwCodecIndex,0,&cchCodecName);
							if(SUCCEEDED(hr))
							{
								pwszCodecName=new WCHAR[cchCodecName];
								if(pwszCodecName)
								{
									hr=pCodecInfo->GetCodecName(WMMEDIATYPE_Audio,dwCodecIndex,pwszCodecName,&cchCodecName);
									if(SUCCEEDED(hr))
									{
										cCombo=WideCharToMultiByte(GetACP(),0,pwszCodecName,cchCodecName,0,0,0,0);
										pszCombo=new char[cCombo];
										if(pszCombo)
										{
											WideCharToMultiByte(GetACP(),0,pwszCodecName,cchCodecName,pszCombo,cCombo,0,0);
											cCombo=SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_ADDSTRING,0,(LPARAM)pszCombo);
											SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_SETITEMDATA,cCombo,dwCodecIndex);
											if(dwCodecIndex==codecindex)
												SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_SETCURSEL,cCombo,0);
											delete [] pszCombo;
										}
									}
									delete [] pwszCodecName;
								}
							}
						}
					}
				}
			}
			pCodecInfo->Release();
		}
		pProfileManager->Release();
	}
	if(SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_GETCURSEL,0,0)==CB_ERR)
		SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_SETCURSEL,0,0);
	curCodec=SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_GETCURSEL,0,0);
	UpdateFormatCombo(hWndDlg,vbr,SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_GETITEMDATA,curCodec,0));
}

void UpdateFormatCombo(HWND hWndDlg, BOOL vbr, DWORD codecindex)
{
	SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_RESETCONTENT,0,0);
	IWMProfileManager* pProfileManager;
	HRESULT hr=WMCreateProfileManager(&pProfileManager);
	if(SUCCEEDED(hr))
	{
		IWMCodecInfo3* pCodecInfo;
		hr=pProfileManager->QueryInterface(IID_IWMCodecInfo3,(void**)&pCodecInfo);
		if(SUCCEEDED(hr))
		{
			IWMMediaProps* pMediaProps;
			WMT_ATTR_DATATYPE datatype=WMT_TYPE_BOOL;
			WCHAR* pwszFormatName=0;
			DWORD cchFormatName=0;
			char* pszCombo=0;
			DWORD cFormats=0;
			DWORD cCombo;
			BOOL bAdd;
			hr=pCodecInfo->SetCodecEnumerationSetting(WMMEDIATYPE_Audio,codecindex,g_wszVBREnabled,datatype,(LPBYTE)&vbr,sizeof(BOOL));
			if(SUCCEEDED(hr))
			{
				hr=pCodecInfo->GetCodecFormatCount(WMMEDIATYPE_Audio,codecindex,&cFormats);
				if(SUCCEEDED(hr)&&cFormats)
				{
					IWMStreamConfig* pStreamConfig;
					for(DWORD index=0;index<cFormats;index++)
					{
						hr=pCodecInfo->GetCodecFormatDesc(WMMEDIATYPE_Audio,codecindex,index,&pStreamConfig,0,&cchFormatName);
						if(SUCCEEDED(hr))
						{
							bAdd=0;
							hr=pStreamConfig->QueryInterface(IID_IWMMediaProps,(void**)&pMediaProps);
							if(SUCCEEDED(hr))
							{
								hr=pMediaProps->GetMediaType(0,&cCombo);
								if(SUCCEEDED(hr)&&cCombo)
								{
									BYTE* pType=new BYTE[cCombo];
									if(pType)
									{
										hr=pMediaProps->GetMediaType((WM_MEDIA_TYPE*)pType,&cCombo);
										if(SUCCEEDED(hr))
										{
											WM_MEDIA_TYPE* pMediaType=(WM_MEDIA_TYPE*)pType;
											WAVEFORMATEX* pwfx=(WAVEFORMATEX*)pMediaType->pbFormat;
											if(pwfx->nChannels<=2)
												bAdd=1;
										}
										delete [] pType;
									}
								}
								pMediaProps->Release();
							}
							pStreamConfig->Release();
							pStreamConfig=0;
							if(bAdd)
							{
								pwszFormatName=new WCHAR[cchFormatName];
								if(pwszFormatName)
								{
									hr=pCodecInfo->GetCodecFormatDesc(WMMEDIATYPE_Audio,codecindex,index,&pStreamConfig,pwszFormatName,&cchFormatName);
									if(SUCCEEDED(hr))
									{
										pStreamConfig->Release();
										pStreamConfig=0;
										if(!wcsstr(pwszFormatName,L"(A/V)"))
										{
											cCombo=WideCharToMultiByte(GetACP(),0,pwszFormatName,cchFormatName,0,0,0,0);
											pszCombo=new char[cCombo];
											if(pszCombo)
											{
												WideCharToMultiByte(GetACP(),0,pwszFormatName,cchFormatName,pszCombo,cCombo,0,0);
												cCombo=SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_ADDSTRING,0,(LPARAM)pszCombo);
												SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_SETITEMDATA,cCombo,index);
												delete [] pszCombo;
											}
										}
									}
									delete [] pwszFormatName;
								}
							}
						}
					}
				}
			}
			pCodecInfo->Release();
		}
		pProfileManager->Release();
	}
	SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_SETCURSEL,0,0);
}

BOOL CALLBACK DialogProc(HWND hWndDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_INITDIALOG:
		{
			SendDlgItemMessage(hWndDlg,IDC_VBR,CB_ADDSTRING,0,(LPARAM)"CBR");
			SendDlgItemMessage(hWndDlg,IDC_VBR,CB_ADDSTRING,0,(LPARAM)"VBR");
			GetConfig();
			curVBR=0;
			if(config&FLAG_VBR)
				curVBR=1;
			SendDlgItemMessage(hWndDlg,IDC_VBR,CB_SETCURSEL,curVBR,0);
			DWORD codecindex=(config>>16)&0xff;
			UpdateCodecCombo(hWndDlg,curVBR,codecindex);
			codecindex=config&0xffff;
			DWORD itemcount=SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_GETCOUNT,0,0);
			for(DWORD index=0;index<itemcount;index++)
			{
				if(SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_GETITEMDATA,index,0)==(long)codecindex)
				{
					SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_SETCURSEL,index,0);
					break;
				}
			}
		}
		return 1;
	case WM_CLOSE:
		PostMessage(hWndDlg,WM_COMMAND,IDCANCEL,0);
		return 1;
	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case IDOK:
				config=SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_GETITEMDATA,curCodec,0);
				config<<=16;
				config+=SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_GETITEMDATA,SendDlgItemMessage(hWndDlg,IDC_FORMAT,CB_GETCURSEL,0,0),0);
				if(curVBR)
					config|=FLAG_VBR;
				SetConfig();
		case IDCANCEL:
			EndDialog(hWndDlg,0);
			return 1;
		case IDC_VBR:
			if(HIWORD(wParam)==CBN_SELCHANGE)
			{
				if(curVBR!=SendDlgItemMessage(hWndDlg,IDC_VBR,CB_GETCURSEL,0,0))
				{
					curVBR=SendDlgItemMessage(hWndDlg,IDC_VBR,CB_GETCURSEL,0,0);
					UpdateCodecCombo(hWndDlg,curVBR,0);
				}
			}
			break;
		case IDC_CODEC:
			if(HIWORD(wParam)==CBN_SELCHANGE)
			{
				if(curCodec!=SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_GETCURSEL,0,0))
				{
					curCodec=SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_GETCURSEL,0,0);
					UpdateFormatCombo(hWndDlg,curVBR,SendDlgItemMessage(hWndDlg,IDC_CODEC,CB_GETITEMDATA,curCodec,0));
				}
			}
		}
 	}
	return 0;
}
