// WavPack encoder plug-in for VUPlayer
// Copyright (C) 2005 James Chapman
// http://www.vuplayer.com

// You will need the WavPack library to build this project.
// You'll also need the Monkey's Audio library for APETag support.

#include <windows.h>
#include <stdio.h>
#include <wavpack/wavpack.h>
#include <ape/all.h>
#include <ape/apetag.h>
#include "../vu_enc.h"
#include "md5.h"
#include "resource.h"

typedef struct
{
    uint32_t bytes_written, first_block_size;
    FILE *file;
    int error;
} write_id;

static int write_block(void *id, void *data, int32_t bcount)
{
    write_id* wid=(write_id*) id;
    if (wid->error)
		return 0;
    if (wid&&wid->file&&data&&bcount)
	{
		if((int32_t)fwrite(data,1,bcount,wid->file)!=bcount)
		{
		    fclose (wid->file);
		    wid->file=0;
			wid->error=1;
			return 0;
		}
		else
		{
			wid->bytes_written+=bcount;
		    if(!wid->first_block_size)
				wid->first_block_size=bcount;
		}
    }
    return 1;
}

#define DEFAULT_QUAL		1
#define DEFAULT_HYBRID		3

#define FLAG_FAST		0x100
#define	FLAG_HIGH		0x200
#define FLAG_HYBRID		0x800

const char description[]="WavPack 4.2";
const int QUALITY_COUNT=3;
const char* QUALITY_STRINGS[QUALITY_COUNT]={"Fast","Normal","High"};
const int HYBRID_COUNT=11;
const int HYBRID_RATES[HYBRID_COUNT]={192,224,256,320,384,448,512,640,768,896,1024};

DWORD config=DEFAULT_HYBRID;
DWORD samplerate=44100;
WORD channels=2;
WORD bps=16;

int32_t wvBuffer[32768];
TAGSTRUCT* tag=0;
WavpackContext* wpc=0;
write_id wvFile;
write_id wvcFile;
md5_state_t md5_context;
char outPath[MAX_PATH];
char wvcPath[MAX_PATH];

/////////////////////
// Internal functions
/////////////////////

void GetConfig()
{
	HKEY hKey;
	if(RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer",0,KEY_QUERY_VALUE,&hKey)==ERROR_SUCCESS)
	{
		BYTE buffer[4];
		DWORD bufSize=4;
		if(RegQueryValueEx(hKey,"encoder_wv",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			config=*(DWORD*)buffer;
		RegCloseKey(hKey);
	}
}

void SetConfig()
{
	DWORD disp;
	HKEY hKey;
	if(RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer",0,0,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,0,&hKey,&disp)==ERROR_SUCCESS)
	{
		RegSetValueEx(hKey,"encoder_wv",0,REG_DWORD,(BYTE*)&config,4);
		RegCloseKey(hKey);
	}
}

BOOL CALLBACK DialogProc(HWND hWndDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_INITDIALOG:
		{
			GetConfig();
			int index;
			char buffer[16]="";
			for(index=0;index<QUALITY_COUNT;index++)
				SendDlgItemMessage(hWndDlg,IDC_QUALITY,CB_ADDSTRING,0,(LPARAM)QUALITY_STRINGS[index]);
			if(config&FLAG_HIGH)
				index=2;
			else
			{
				if(config&FLAG_FAST)
					index=0;
				else
					index=1;
			}
			SendDlgItemMessage(hWndDlg,IDC_QUALITY,CB_SETCURSEL,index,0);
			for(index=0;index<HYBRID_COUNT;index++)
			{
				itoa(HYBRID_RATES[index],buffer,10);
				strcat(buffer," kbps");
				SendDlgItemMessage(hWndDlg,IDC_BITRATE,CB_ADDSTRING,0,(LPARAM)buffer);
			}
			index=(BYTE)config;
			SendDlgItemMessage(hWndDlg,IDC_BITRATE,CB_SETCURSEL,index,0);
			if(config&FLAG_HYBRID)
			{
				SendDlgItemMessage(hWndDlg,IDC_RADIO2,BM_SETCHECK,BST_CHECKED,0);
				EnableWindow(GetDlgItem(hWndDlg,IDC_BITRATE),TRUE);
			}
			else
			{
				SendDlgItemMessage(hWndDlg,IDC_RADIO1,BM_SETCHECK,BST_CHECKED,0);
				EnableWindow(GetDlgItem(hWndDlg,IDC_BITRATE),FALSE);
			}
		}
		return 1;
	case WM_CLOSE:
		PostMessage(hWndDlg,WM_COMMAND,IDCANCEL,0);
		return 1;
	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case IDOK:
			config=SendDlgItemMessage(hWndDlg,IDC_BITRATE,CB_GETCURSEL,0,0);
			if(SendDlgItemMessage(hWndDlg,IDC_QUALITY,CB_GETCURSEL,0,0)==2)
				config|=FLAG_HIGH;
			else
			{
				if(!SendDlgItemMessage(hWndDlg,IDC_QUALITY,CB_GETCURSEL,0,0))
					config|=FLAG_FAST;
			}
			if(IsWindowEnabled(GetDlgItem(hWndDlg,IDC_BITRATE)))
				config|=FLAG_HYBRID;
			SetConfig();
		case IDCANCEL:
			EndDialog(hWndDlg,0);
			return 1;
		case IDC_RADIO1:
			if(SendDlgItemMessage(hWndDlg,IDC_RADIO1,BM_GETCHECK,0,0)==BST_CHECKED)
				EnableWindow(GetDlgItem(hWndDlg,IDC_BITRATE),FALSE);
			else
				EnableWindow(GetDlgItem(hWndDlg,IDC_BITRATE),TRUE);
			break;
		case IDC_RADIO2:
			if(SendDlgItemMessage(hWndDlg,IDC_RADIO2,BM_GETCHECK,0,0)==BST_CHECKED)
				EnableWindow(GetDlgItem(hWndDlg,IDC_BITRATE),TRUE);
			else
				EnableWindow(GetDlgItem(hWndDlg,IDC_BITRATE),FALSE);
		}
 	}
	return 0;
}

/////////////////////
// Exported functions
/////////////////////

LPCTSTR QueryEncoder()
{
	return description;
}

LPCTSTR OpenEx(LPCTSTR filename, TAGSTRUCT* pTag, DWORD dwLength, DWORD freq, WORD chan, WORD fmt)
{
	if(fmt!=FORMAT_16BIT)
		return 0;
	samplerate=freq;
	channels=chan;
	return Open(filename,pTag,dwLength);
}

LPCTSTR Open(LPCTSTR filename, TAGSTRUCT* pTag, DWORD dwLength)
{
	ZeroMemory(&wvFile,sizeof(write_id));
	ZeroMemory(&wvcFile,sizeof(write_id));
	wpc=WavpackOpenFileOutput(write_block,&wvFile,&wvcFile);
	GetConfig();
	tag=pTag;
	if(wpc)
	{
		ZeroMemory(outPath,MAX_PATH);
		strcpy(outPath,filename);
		strcat(outPath,".wv");
		wvFile.file=fopen(outPath,"w+b");
		if(wvFile.file)
		{
			WavpackConfig wvconfig;
			ZeroMemory(&wvconfig,sizeof(WavpackConfig));
			wvconfig.flags|=CONFIG_MD5_CHECKSUM;
			if(config&FLAG_HIGH)
				wvconfig.flags|=CONFIG_HIGH_FLAG;
			else
			{
				if(config&FLAG_FAST)
					wvconfig.flags|=CONFIG_FAST_FLAG;
			}
			if(config&FLAG_HYBRID)
			{
				BYTE i=(BYTE)config;
				if(i<HYBRID_COUNT)
				{
					wvconfig.flags|=CONFIG_HYBRID_FLAG;
					wvconfig.flags|=CONFIG_BITRATE_KBPS;
					wvconfig.flags|=CONFIG_CREATE_WVC;
					wvconfig.bitrate=(float)HYBRID_RATES[i];
				}
			}
			BOOL success=1;
			if(wvconfig.flags&CONFIG_CREATE_WVC)
			{
				strcpy(wvcPath,outPath);
				strcat(wvcPath,"c");
				wvcFile.file=fopen(wvcPath,"w+b");
				if(!wvcFile.file)
					success=0;
			}
			if(success)
			{
				wvconfig.sample_rate=samplerate;
				wvconfig.num_channels=channels;
				wvconfig.bits_per_sample=16;
				wvconfig.bytes_per_sample=2;
				wvconfig.channel_mask=0x05-channels;
				WavpackSetConfiguration(wpc,&wvconfig,dwLength/channels/2);
			
				RiffChunkHeader riffhdr;
				ChunkHeader datahdr;
				ChunkHeader fmthdr;
				WaveHeader wavhdr;
				strncpy(riffhdr.ckID,"RIFF",4);
				riffhdr.ckSize=sizeof(riffhdr)+16+sizeof(datahdr)+dwLength;
				strncpy(riffhdr.formType,"WAVE",4);

				strncpy(fmthdr.ckID,"fmt ",4);
				fmthdr.ckSize=16;

				wavhdr.FormatTag=1;
				wavhdr.NumChannels=channels;
				wavhdr.SampleRate=samplerate;
				wavhdr.BytesPerSecond=samplerate*channels*2;
				wavhdr.BlockAlign=channels*2;
				wavhdr.BitsPerSample=16;	

				strncpy(datahdr.ckID,"data",4);
				datahdr.ckSize=dwLength; 

			    if(WavpackAddWrapper(wpc,&riffhdr,sizeof(riffhdr))&&
					WavpackAddWrapper(wpc,&fmthdr,sizeof(fmthdr))&&
					WavpackAddWrapper(wpc,&wavhdr,fmthdr.ckSize)&&
					WavpackAddWrapper(wpc,&datahdr,sizeof(datahdr)))
				{
					if(WavpackPackInit(wpc))
					{
						md5_init(&md5_context);
						return outPath;
					}
				}
			}
			if(wvFile.file)
			{
				fclose(wvFile.file);
				wvFile.file=0;
			}
			if(wvcFile.file)
			{
				fclose(wvcFile.file);
				wvcFile.file=0;
			}
		}
		WavpackCloseFile(wpc);
	}
	return 0;
}

void Close()
{
	if(wvFile.file)
	{
		WavpackFlushSamples(wpc);
		md5_byte_t md5_digest[16];
		md5_finish(&md5_context,md5_digest);
		WavpackStoreMD5Sum(wpc,md5_digest);
		WavpackFlushSamples(wpc);
		if(WavpackGetSampleIndex(wpc)!=WavpackGetNumSamples(wpc))
		{
			char* block_buff=new char[wvFile.first_block_size];
			if(block_buff)
			{
				fseek(wvFile.file,0,SEEK_SET);
				fread(block_buff,wvFile.first_block_size,1,wvFile.file);
				WavpackUpdateNumSamples(wpc,block_buff);
				if(WavpackGetWrapperLocation(block_buff))
				{
					uint32_t datasize=WavpackGetSampleIndex(wpc)*WavpackGetNumChannels(wpc)*WavpackGetBytesPerSample(wpc);
					uint32_t wrappersize=36;
					RiffChunkHeader* riffhdr=(RiffChunkHeader*)WavpackGetWrapperLocation(block_buff);
					ChunkHeader* datahdr=(ChunkHeader*)((char*)riffhdr+wrappersize);
					if(!strncmp(riffhdr->ckID,"RIFF",4))
						riffhdr->ckSize=wrappersize+datasize;
					if(!strncmp(datahdr->ckID,"data",4))
						datahdr->ckSize=datasize;
				}
				fseek(wvFile.file,0,SEEK_SET);
				fwrite(block_buff,wvFile.first_block_size,1,wvFile.file);
				delete [] block_buff;
			}
			if(wvcFile.file)
			{
				block_buff=new char[wvcFile.first_block_size];
				if(block_buff)
				{
					fseek(wvcFile.file,0,SEEK_SET);
					fread(block_buff,wvcFile.first_block_size,1,wvcFile.file);
					WavpackUpdateNumSamples(wpc,block_buff);
					fseek(wvcFile.file,0,SEEK_SET);
					fwrite(block_buff,wvcFile.first_block_size,1,wvcFile.file);
					delete [] block_buff;
				}
			}
		}
		WavpackCloseFile(wpc);
		fclose(wvFile.file);
		int wLen=MultiByteToWideChar(GetACP(),0,outPath,-1,0,0);
		if(wLen)
		{
			WCHAR* wBuf=new WCHAR[wLen];
			if(wBuf)
			{
				MultiByteToWideChar(GetACP(),0,outPath,-1,wBuf,wLen);
				CAPETag apetag(wBuf);
				apetag.SetFieldString(APE_TAG_FIELD_TITLE,tag->title,1);
				apetag.SetFieldString(APE_TAG_FIELD_ARTIST,tag->artist,1);
				apetag.SetFieldString(APE_TAG_FIELD_ALBUM,tag->album,1);
				apetag.SetFieldString(APE_TAG_FIELD_GENRE,tag->genre,1);
				apetag.SetFieldString(APE_TAG_FIELD_YEAR,tag->year,1);
				apetag.SetFieldString(APE_TAG_FIELD_COMMENT,tag->comment,1);
				char tracknum[32]="";
				itoa(tag->tracknum,tracknum,10);
				apetag.SetFieldString(APE_TAG_FIELD_TRACK,tracknum,1);
				apetag.Save();
				delete [] wBuf;
			}
		}
	}
	if(wvcFile.file)
	{
		fclose(wvcFile.file);
		if(!wvcFile.bytes_written)
			DeleteFile(wvcPath);
	}
}

BOOL Write(BYTE* pData, DWORD dwSize)
{
	short* pShort=(short*)pData;
	for(DWORD i=0;i<dwSize/2;i++)
		wvBuffer[i]=pShort[i];
	md5_append(&md5_context,pData,dwSize);
	return WavpackPackSamples(wpc,wvBuffer,dwSize/channels/2);
}

void Configure(HINSTANCE hInst, HWND hWnd)
{
	DialogBox(hInst,MAKEINTRESOURCE(IDD_CONFIG),hWnd,DialogProc);
}

